# -*- coding: utf-8 -*-
"""Lispython: the love child of Python and Scheme.

Powered by Pydialect and unpythonic."""

from macropy.core.quotes import macros, q, name

from dialects.util import splice_ast

def ast_transformer(module_body):
    with q as template:
        from unpythonic.syntax import macros, tco, autoreturn, \
                                      multilambda, quicklambda, namedlambda, \
                                      let, letseq, letrec, do, do0, \
                                      dlet, dletseq, dletrec, \
                                      blet, bletseq, bletrec, \
                                      let_syntax, abbrev, \
                                      cond
        # auxiliary syntax elements for the macros
        from unpythonic.syntax import local, delete, where, block, expr, f, _
        from unpythonic import cons, car, cdr, ll, llist, nil, prod, dyn
        with namedlambda:  # MacroPy #21 (nontrivial two-pass macro; seems I didn't get the fix right)
            with autoreturn, quicklambda, multilambda, tco:
                name["__paste_here__"]
    return splice_ast(module_body, template, "__paste_here__")

def rejoice():
    """**Schemers rejoice!**::

    Multiple musings mix in a lambda,
    Lament no longer the lack of let.
    Languish no longer labelless, lambda,
    Linked lists cons and fold.
    Tail-call into recursion divine,
    The final value always provide."""
    return rejoice.__doc__



# -*- coding: utf-8 -*-

__all__ = ["splice_ast"]

from ast import Expr, Name, If, Num, ImportFrom, Pass, copy_location

try:
    from macropy.core.walkers import Walker
except ImportError:
    Walker = None

def splice_ast(body, template, tag):
    """In an AST transformer, splice module body into template.

    Imports for MacroPy macros are handled specially, gathering them all at the
    front, so that MacroPy sees them. Any macro imports in the template are
    placed first (in the order they appear in the template), followed by any
    macro imports in the user code (in the order they appear in the user code).

    This function is provided as a convenience for modules that define dialects.
    We use MacroPy to perform the splicing, so this function is only available
    when MacroPy is installed (``ImportError`` is raised if not). Installation
    status is checked only once per session, when ``dialects.util`` is first
    imported.

    Parameters:

        ``body``: ``list`` of statements
            Module body of the original user code (input).

        ``template``: ``list`` of statements
            Template for the module body of the new module (output).

            Must contain a marker that indicates where ``body`` is to be
            spliced in. The marker is an ``ast.Name`` node whose ``id``
            attribute matches the value of the ``tag`` string.

        ``tag``: ``str``
            The value of the ``id`` attribute of the marker in ``template``.

    Returns the new module body, i.e. ``template`` with ``body`` spliced in.

    Example::

        marker = q[name["__paste_here__"]]      # MacroPy, or...
        marker = ast.Name(id="__paste_here__")  # ...manually

        ...  # create template, place the marker in it

        dialects.util.splice_ast(body, template, "__paste_here__")

    """
    if not Walker:  # optional dependency for Pydialect, but mandatory for this util
        raise ImportError("macropy.core.walkers.Walker not found; MacroPy likely not installed")
    if not body:  # ImportError because this occurs during the loading of a module written in a dialect.
        raise ImportError("expected at least one statement or expression in module body")

    def is_paste_here(tree):
        return type(tree) is Expr and type(tree.value) is Name and tree.value.id == tag
    def is_macro_import(tree):
        return type(tree) is ImportFrom and tree.names[0].name == "macros"

    # XXX: MacroPy's debug logger will sometimes crash if a node is missing a source location.
    # In general, dialect templates are fully macro-generated with no source location info to start with.
    # Pretend it's all at the start of the user module.
    locref = body[0]
    @Walker
    def fix_template_srcloc(tree, **kw):
        if not all(hasattr(tree, x) for x in ("lineno", "col_offset")):
            tree = copy_location(tree, locref)
        return tree

    @Walker
    def extract_macro_imports(tree, *, collect, **kw):
        if is_macro_import(tree):
            collect(tree)
            tree = copy_location(Pass(), tree)  # must output a node so replace by a pass stmt
        return tree

    template = fix_template_srcloc.recurse(template)
    template, template_macro_imports = extract_macro_imports.recurse_collect(template)
    body, user_macro_imports = extract_macro_imports.recurse_collect(body)

    @Walker
    def splice_body_into_template(tree, *, stop, **kw):
        if not is_paste_here(tree):
            return tree
        stop()  # prevent infinite recursion in case the user code contains a Name that looks like the marker
        return If(test=Num(n=1),
                  body=body,
                  orelse=[],
                  lineno=tree.lineno, col_offset=tree.col_offset)
    finalbody = splice_body_into_template.recurse(template)
    return template_macro_imports + user_macro_imports + finalbody



# -*- coding: utf-8 -*-
"""Pytkell: Python with automatic currying and lazy functions.

Powered by Pydialect and unpythonic."""

from macropy.core.quotes import macros, q, name

from dialects.util import splice_ast

def ast_transformer(module_body):
    with q as template:
        from macropy.quick_lambda import macros, lazy
        from unpythonic.syntax import macros, lazify, lazyrec, curry, \
                                      let, letseq, letrec, do, do0, \
                                      dlet, dletseq, dletrec, \
                                      blet, bletseq, bletrec, \
                                      cond, forall
        # auxiliary syntax elements for the macros
        from unpythonic.syntax import local, delete, where, insist, deny
        # functions that have a haskelly feel to them
        from unpythonic import foldl, foldr, scanl, scanr, \
                               s, m, mg, frozendict, \
                               memoize, fupdate, fup, \
                               gmemoize, imemoize, fimemoize, \
                               islice, take, drop, split_at, first, second, nth, last, \
                               flip, rotate
        from unpythonic import composerc as compose  # compose from Right, Currying (Haskell's . operator)
        # this is a bit lispy, but we're not going out of our way to provide
        # a haskelly surface syntax for these.
        from unpythonic import cons, car, cdr, ll, llist, nil
        with curry, lazify:
            name["__paste_here__"]
    return splice_ast(module_body, template, "__paste_here__")



# -*- coding: utf-8 -*-
"""Importer for Python dialects."""

__all__ = ["DialectFinder"]

import ast
import importlib
from importlib.util import spec_from_loader
import logging
import sys
import re

try:
    import macropy.core
except ImportError:
    macropy = None

logger = logging.getLogger(__name__)

# This is essentially a copy of ``macropy.core.import_hooks.MacroLoader``, copied to
# make sure that the implementation won't go out of sync with our ``DialectFinder``.
# The export machinery has been removed as unnecessary for language experimentation;
# but is trivial to add back if needed (see the sources of MacroPy 1.1.0b2).
class DialectLoader:
    def __init__(self, nomacro_spec, code, tree):
        self.nomacro_spec = nomacro_spec
        self.code = code
        self.tree = tree

    def create_module(self, spec):
        pass

    def exec_module(self, module):
        exec(self.code, module.__dict__)

    def get_filename(self, fullname):
        return self.nomacro_spec.loader.get_filename(fullname)

    def is_package(self, fullname):
        return self.nomacro_spec.loader.is_package(fullname)

# barebones unpythonic.misc.call but let's not depend on a library we don't otherwise need
def singleton(cls):
    return cls()

# Like MacroPy's MacroFinder:
#
# - This is a meta_path finder, because we're looking for nonstandard things
#   inside standard-ish Python modules that can be found/loaded using the
#   standard mechanisms.
#
# - Macro expansion is performed already at the finder step, because in order to
#   detect whether to use the macro loader, the input must be scanned for macros.
#   We could dispatch to a custom loader immediately after detecting that the
#   module uses a dialect, but have chosen to just inherit this design.

@singleton
class DialectFinder:
    """Importer that matches any module that has a 'from __lang__ import xxx'."""

    def _find_spec_nomacro(self, fullname, path, target=None):
        """Try to find the original, non macro-expanded module using all the
        remaining meta_path finders (except MacroPy's, to avoid handling
        macros twice).
        """
        spec = None
        for finder in sys.meta_path:
            # when testing with pytest, it installs a finder that for
            # some yet unknown reasons makes macros expansion
            # fail. For now it will just avoid using it and pass to
            # the next one
            if finder is self or (macropy and finder is macropy.core.import_hooks.MacroFinder) or \
               'pytest' in finder.__module__:
                continue
            if hasattr(finder, 'find_spec'):
                spec = finder.find_spec(fullname, path, target=target)
            elif hasattr(finder, 'load_module'):
                spec = spec_from_loader(fullname, finder)
            if spec is not None:
                break
        return spec

    def expand_macros(self, source_code, filename, fullname, spec, lang_module):
        """Parse, apply AST transforms, and compile.

        Parses the source_code, applies the ast_transformer of the dialect,
        and macro-expands the resulting AST if it has macros. Then compiles
        the final AST.

        Returns both the compiled new AST, and the raw new AST.
        """
        logger.info('Parse in file {} (module {})'.format(filename, fullname))
        tree = ast.parse(source_code)

        if not (isinstance(tree, ast.Module) and tree.body):
            msg = "Expected a Module node with at least one statement or expression in file {} (module {})".format(filename, fullname)
            logger.error(msg)
            raise SyntaxError(msg)

        # The 'from __lang__ import xxx' has done its job, replace it with
        # '__lang__ = "xxx"' for run-time introspection.
        def isimportdialect(tree):
            return type(tree) is ast.ImportFrom and tree.module == "__lang__"
        # q[] would be much easier but we want to keep MacroPy optional.
        # We need to be careful to insert a location to each AST node
        # in case we're not invoking MacroPy (which fixes missing locations).
        def make_langname_setter(tree):
            s = ast.Str(s=tree.names[0].name)
            s = ast.copy_location(s, tree)
            n = ast.Name(id="__lang__", ctx=ast.Store())
            n = ast.copy_location(n, tree)
            a = ast.Assign(targets=[n], value=s)
            a = ast.copy_location(a, tree)
            return a

        if isimportdialect(tree.body[0]):
            preamble = [make_langname_setter(tree.body[0])]
            thebody = tree.body[1:]
        # variant with module docstring
        elif len(tree.body) > 1 and isinstance(tree.body[0], ast.Expr) and \
             isimportdialect(tree.body[1]):
            preamble = [tree.body[0], make_langname_setter(tree.body[1])]
            thebody = tree.body[2:]
        else:  # we know the lang-import is there; it's in some other position
            msg = "Misplaced lang-import in file {} (module {})".format(filename, fullname)
            logger.error(msg)
            raise SyntaxError(msg)

        if hasattr(lang_module, "ast_transformer"):
            logger.info('Dialect AST transform in file {} (module {})'.format(filename, fullname))
            thebody = lang_module.ast_transformer(thebody)
        tree.body = preamble + thebody

        # detect macros **after** any dialect-level whole-module transform
        new_tree = tree
        if macropy:
            logger.info('Detect macros in file {} (module {})'.format(filename, fullname))
            bindings = macropy.core.macros.detect_macros(tree, spec.name,
                                                         spec.parent,
                                                         spec.name)
            if bindings:  # expand macros
                logger.info('Expand macros in file {} (module {})'.format(filename, fullname))
                modules = []
                for mod, bind in bindings:
                    modules.append((importlib.import_module(mod), bind))
                new_tree = macropy.core.macros.ModuleExpansionContext(
                    tree, source_code, modules).expand_macros()

        try:
            # MacroPy uses the old tree here as input to compile(), but it doesn't matter,
            # since ``ModuleExpansionContext.expand_macros`` mutates the tree in-place.
            logger.info('Compile file {} (module {})'.format(filename, fullname))
            return compile(new_tree, filename, "exec"), new_tree
        except Exception:
            logger.error("Error while compiling file {} (module {})".format(filename, fullname))
            raise

    def find_spec(self, fullname, path, target=None):
        spec = self._find_spec_nomacro(fullname, path, target)
        if spec is None or not (hasattr(spec.loader, 'get_source') and
                                callable(spec.loader.get_source)):  # noqa: E128
            if fullname != 'org':
                # stdlib pickle.py at line 94 contains a ``from
                # org.python.core for Jython which is always failing,
                # of course
                logger.debug('Failed finding spec for {}'.format(fullname))
            return
        origin = spec.origin
        if origin == 'builtin':
            return
        try:
            source = spec.loader.get_source(fullname)
        except ImportError:
            logger.debug('Loader for {} was unable to find the sources'.format(fullname))
            return
        except Exception:
            logger.error('Loader for {} raised an error'.format(fullname))
            return
        if not source:  # some loaders may return None for the sources, without raising an exception
            logger.debug('Loader returned empty sources for {}'.format(fullname))
            return

        lang_import = "from __lang__ import"
        if lang_import not in source:
            return  # this module does not use a dialect

        # Detect the dialect... ugh!
        #   - At this point, the input is text.
        #   - It's not parseable by ast.parse, because a dialect may introduce
        #     new surface syntax.
        #   - Similarly, it's not tokenizable by stdlib's tokenizer, because
        #     a dialect may customize what constitutes a token.
        #   - So we can only rely on the literal text "from __lang__ import xxx".
        #   - This is rather similar to how Racket heavily constrains what may
        #     appear on the #lang line.
        matches = re.findall(r"from __lang__ import\s+([0-9a-zA-Z_]+)\s*$", source, re.MULTILINE)
        if len(matches) != 1:
            msg = "Expected exactly one lang-import with one dialect name"
            logger.error(msg)
            raise SyntaxError(msg)
        dialect_name = matches[0]

        try:
            logger.info("Detected dialect '{}' in module '{}', loading dialect".format(dialect_name, fullname))
            lang_module = importlib.import_module(dialect_name)
        except ImportError as err:
            msg = "Could not import dialect module '{}'".format(dialect_name)
            logger.error(msg)
            raise ImportError(msg) from err
        if not any(hasattr(lang_module, x) for x in ("source_transformer", "ast_transformer")):
            msg = "Module '{}' has no dialect transformers".format(dialect_name)
            logger.error(msg)
            raise ImportError(msg)

        if hasattr(lang_module, "source_transformer"):
            logger.info('Dialect source transform in {}'.format(fullname))
            source = lang_module.source_transformer(source)
            if not source:
                msg = "Empty source text after dialect source transform in {}".format(fullname)
                logger.error(msg)
                raise SyntaxError(msg)
            if lang_import not in source:  # preserve invariant
                msg = 'Dialect source transform for {} should not delete the lang-import'.format(fullname)
                logger.error(msg)
                raise RuntimeError(msg)

        code, tree = self.expand_macros(source, origin, fullname, spec, lang_module)

        # Unlike macropy.core.import_hooks.MacroLoader, which exits at this point if there
        # were no macros, we always process the module (because it was explicitly tagged
        # as a dialect, and pure source-transform dialects are also allowed).

        loader = DialectLoader(spec, code, tree)
        return spec_from_loader(fullname, loader)



# -*- coding: utf-8 -*-
"""LisThEll: it's not Lisp, it's not Python, it's not Haskell.

Powered by Pydialect and unpythonic."""

from macropy.core.quotes import macros, q, name

from dialects.util import splice_ast

def ast_transformer(module_body):
    with q as template:
        from unpythonic.syntax import macros, prefix, curry
        # auxiliary syntax elements for the macros
        from unpythonic.syntax import q, u, kw
        from unpythonic import apply
        from unpythonic import composerc as compose  # compose from Right, Currying
        with prefix, curry:
            name["__paste_here__"]
    return splice_ast(module_body, template, "__paste_here__")



# -*- coding: utf-8 -*-
#
"""setuptools-based setup.py for Pydialect.

Tested on Python 3.4.

Usage as usual with setuptools:
    python3 setup.py build
    python3 setup.py sdist
    python3 setup.py bdist_wheel --universal
    python3 setup.py install

For details, see
    http://setuptools.readthedocs.io/en/latest/setuptools.html#command-reference
or
    python3 setup.py --help
    python3 setup.py --help-commands
    python3 setup.py --help bdist_wheel  # or any command
"""

#########################################################
# General config
#########################################################

# Name of the top-level package of the library.
#
# This is also the top level of its source tree, relative to the top-level project directory setup.py resides in.
#
libname="pydialect"

# Short description for package list on PyPI
#
SHORTDESC="Build languages on Python."

# Long description for package homepage on PyPI
#
DESC="""Pydialect makes Python into a language platform, à la Racket. It provides the
plumbing that allows to create, in Python, dialects that compile into Python
at import time.

An extension to the Python language doesn't need to make it into the Python core,
*or even be desirable for inclusion* into the Python core, in order to be useful.

Building on functions and syntactic macros, customization of the language itself
is one more tool for the programmer to extract patterns, at a higher level.
Hence, beside language experimentation, such extensions can be used as a
framework that allows shorter and/or more readable programs.

Pydialect places language-creation power in the hands of its users, without the
need to go to extreme lengths to hack CPython itself or implement from scratch
a custom language that compiles to Python AST or bytecode.

Pydialect is geared toward creating languages that extend Python and look
almost like Python, but extend or modify its syntax and/or semantics.
Hence *dialects*.

That said, Pydialect itself is only a lightweight infrastructure hook that makes
it convenient to define and use dialects. To implement the actual semantics
for your dialect (which is where all the interesting things happen), you may
want to look at [MacroPy](https://github.com/azazel75/macropy). Examples can be
found in [unpythonic](https://github.com/Technologicat/unpythonic); see especially
the macros (comprising about one half of ``unpythonic``). On packaging a set of
semantics into a Pydialect definition, look at the example dialects included
in the Pydialect distribution.

Example of a module using a dialect::

    from __lang__ import lispython

    print("hello, my dialect is {}".format(__lang__))

    c = cons(1, 2)
    assert tuple(c) == (1, 2)
    assert car(c) == 1
    assert cdr(c) == 2
    assert ll(1, 2, 3) == llist((1, 2, 3))

    x = let[(a, 21) in 2*a]
    assert x == 42

    x = letseq[((a, 1),
                (a, 2*a),
                (a, 2*a)) in
               a]
    assert x == 4

    a = lambda x: cond[x < 0, "nope",
                       x % 2 == 0, "even",
                       "odd"]
    assert a(-1) == "nope"
    assert a(2) == "even"
    assert a(3) == "odd"

    def fact(n):
        def f(k, acc):
            if k == 1:
                return acc
            f(k - 1, k*acc)  # implicit return in tail position, like in Lisps
        f(n, acc=1)
    assert fact(4) == 24
    fact(5000)  # automatic TCO, no crash

    square = lambda x: x**2
    assert square(3) == 9
    assert square.__name__ == "square"  # lambdas are auto-named
"""

# Set up data files for packaging.
#
# Directories (relative to the top-level directory where setup.py resides) in which to look for data files.
datadirs  = ()

# File extensions to be considered as data files. (Literal, no wildcards.)
dataexts  = (".py", ".ipynb",  ".sh",  ".lyx", ".tex", ".txt", ".pdf")

# Standard documentation to detect (and package if it exists).
#
standard_docs     = ["README", "LICENSE", "TODO", "CHANGELOG", "AUTHORS"]  # just the basename without file extension
standard_doc_exts = [".md", ".rst", ".txt", ""]  # commonly .md for GitHub projects, but other projects may use .rst or .txt (or even blank).

#########################################################
# Init
#########################################################

import os
from setuptools import setup

# Gather user-defined data files
#
# http://stackoverflow.com/questions/13628979/setuptools-how-to-make-package-contain-extra-data-folder-and-all-folders-inside
#
datafiles = []
#getext = lambda filename: os.path.splitext(filename)[1]
#for datadir in datadirs:
#    datafiles.extend( [(root, [os.path.join(root, f) for f in files if getext(f) in dataexts])
#                       for root, dirs, files in os.walk(datadir)] )

# Add standard documentation (README et al.), if any, to data files
#
detected_docs = []
for docname in standard_docs:
    for ext in standard_doc_exts:
        filename = "".join( (docname, ext) )  # relative to the directory in which setup.py resides
        if os.path.isfile(filename):
            detected_docs.append(filename)
datafiles.append( ('.', detected_docs) )

# Extract __version__ from the package __init__.py
# (since it's not a good idea to actually run __init__.py during the build process).
#
# http://stackoverflow.com/questions/2058802/how-can-i-get-the-version-defined-in-setup-py-setuptools-in-my-package
#
import ast
init_py_path = os.path.join('dialects', '__init__.py')
version = '0.0.unknown'
try:
    with open(init_py_path) as f:
        for line in f:
            if line.startswith('__version__'):
                version = ast.parse(line).body[0].value.s
                break
        else:
            print( "WARNING: Version information not found in '%s', using placeholder '%s'" % (init_py_path, version), file=sys.stderr )
except FileNotFoundError:
    print( "WARNING: Could not find file '%s', using placeholder version information '%s'" % (init_py_path, version), file=sys.stderr )

#########################################################
# Call setup()
#########################################################

setup(
    name = "pydialect",
    version = version,
    author = "Juha Jeronen",
    author_email = "juha.m.jeronen@gmail.com",
    url = "https://github.com/Technologicat/pydialect",

    description = SHORTDESC,
    long_description = DESC,

    license = "BSD",

    # free-form text field; http://stackoverflow.com/questions/34994130/what-platforms-argument-to-setup-in-setup-py-does
    platforms = ["Linux"],

    # See
    #    https://pypi.python.org/pypi?%3Aaction=list_classifiers
    #
    # for the standard classifiers.
    #
    classifiers = [ "Development Status :: 4 - Beta",
                    "Environment :: Console",
                    "Intended Audience :: Developers",
                    "License :: OSI Approved :: BSD License",
                    "Operating System :: POSIX :: Linux",
                    "Programming Language :: Python",
                    "Programming Language :: Python :: 3",
                    "Programming Language :: Python :: 3.4",
                    "Topic :: Software Development :: Libraries",
                    "Topic :: Software Development :: Libraries :: Python Modules"
                  ],

    # See
    #    http://setuptools.readthedocs.io/en/latest/setuptools.html
    #
    setup_requires = [],
    install_requires = [],
    # Lispython is for production, the rest of the example dialects are just for testing.
    provides = ["pydialect", "dialects", "lispython"],

    # keywords for PyPI (in case you upload your project)
    #
    # e.g. the keywords your project uses as topics on GitHub, minus "python" (if there)
    #
    keywords = ["metaprogramming", "programming-language-development", "dialects"],

    # Declare packages so that  python -m setup build  will copy .py files (especially __init__.py).
    #
    # This **does not** automatically recurse into subpackages, so they must also be declared.
    #
    packages = ["dialects", "lispython", "pytkell", "listhell"],

    scripts = ["pydialect"],

    zip_safe = False,  # macros are not zip safe, because the zip importer fails to find sources, and MacroPy needs them.

    # Custom data files not inside a Python package
    data_files = datafiles
)




# -*- coding: utf-8 -*-
"""Install the Pydialect import hook.

Reloading this module refreshes the import hook to the start of ``sys.meta_path``.
"""

from . import importer
import sys

def _find_dialectfinder():
    cls = type(importer.DialectFinder)  # singleton, but id might change if module reloaded
    for j, finder in enumerate(sys.meta_path):
        if isinstance(finder, cls):
            return j
    return -1

j = _find_dialectfinder()
if j > -1:
    sys.meta_path.pop(j)

sys.meta_path.insert(0, importer.DialectFinder)
